import os
from enum import Enum, auto, unique
from typing import Union

import grpc
from jetpack._remote import codec
from jetpack.config import namespace, symbols
from jetpack.proto.runtime.v1alpha1 import remote_pb2, remote_pb2_grpc


# assigns human-readable values, as in: LaunchJobMode.FIRE_AND_FORGET = 'FIRE_AND_FORGET'
# https://docs.python.org/3/library/enum.html#using-automatic-values
class AutoName(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name


@unique
class LaunchJobMode(AutoName):
    ASYNC = auto()
    ASYNC_FIRE_AND_FORGET = auto()
    FIRE_AND_FORGET = auto()
    BLOCKING = auto()

    def is_async(self):
        return self == self.ASYNC or self == self.ASYNC_FIRE_AND_FORGET

    def is_fire_and_forget(self):
        return self == self.FIRE_AND_FORGET or self == self.ASYNC_FIRE_AND_FORGET

    # Named this is_wait_for_response instead of is_blocking because the latter
    # sounds confusing with async. Maybe we should rename BLOCKING to REGULAR?
    def is_wait_for_response(self):
        return self == self.ASYNC or self == self.BLOCKING


class JetpackException(Exception):
    """Base class for exceptions in this module"""

    pass


class RuntimeException(JetpackException):
    """Exception raised for errors in the Jetpack runtime and kubernetes."""

    def __init__(self, message):
        self.message = message


class ApplicationException(JetpackException):
    """Exception raised for errors from application-code that is using the SDK.

    TODO DEV-157
    For exceptions raised by remote functions and jobs, we serialize the
    userland exception in the backend and save it here. The userland exception
    is re-raised by the SDK for the caller of the remote function or job.
    """

    def __init__(self, message):
        self.message = message


class NoControllerAddressError(JetpackException):
    pass


class Client:
    def __init__(self):
        host = os.environ.get(
            "JETPACK_CONTROLLER_HOST",
            "remotesvc.jetpack-runtime.svc.cluster.local",
        )
        port = os.environ.get("JETPACK_CONTROLLER_PORT", "443")
        self.address = f"{host.strip()}:{port.strip()}"
        self.stub = None
        self.async_stub = None
        self.is_dialed = False  # TODO: Mutex needed?
        self.dry_run = False
        self.dry_run_last_request = None

    def dial(self):
        if not self.address:
            raise NoControllerAddressError("Controller address is not set")
        # Since this is inter-cluster communication, insecure is fine.
        # In the future this won't even leave the pod, and use a sidecar so
        # it will be localhost.

        # TODO(Landau): When/how should we close the channels?
        channel = grpc.insecure_channel(self.address)
        self.stub = remote_pb2_grpc.RemoteExecutorStub(channel)
        async_channel = grpc.aio.insecure_channel(self.address)
        self.async_stub = remote_pb2_grpc.RemoteExecutorStub(async_channel)

        self.is_dialed = True

    def launch_job(self, qualified_name: str, module_id: str, args=None, kwargs=None):
        """Launches a k8s job and blocks until the job completes, before returning.
        For now this function assumes job will live in same namespace where the
        launcher is located.

        Keyword arguments:
        qualified_name -- qualified name as produced by utils.qualified_func_name
        module -- jetpack module where the job resides. Used to determine correct
        docker image.
        """
        response = self._launch_job(
            LaunchJobMode.BLOCKING, qualified_name, module_id, args, kwargs
        )

        return self._transform_response_exception(response)

    async def launch_async_job(
        self, qualified_name: str, module_id: str, args=None, kwargs=None
    ):
        response = await self._launch_job(
            LaunchJobMode.ASYNC, qualified_name, module_id, args, kwargs
        )

        return self._transform_response_exception(response)

    def launch_fire_and_forget_job(
        self, qualified_name: str, module_id: str, args=None, kwargs=None
    ):
        """Launches a k8s job. For now this function assumes job will live in
        same namespace where the launcher is located.

        Keyword arguments:
        qualified_name -- qualified name as produced by utils.qualified_func_name
        module -- jetpack module where the job resides. Used to determine correct
        docker image.
        """
        return self._launch_job(
            LaunchJobMode.FIRE_AND_FORGET, qualified_name, module_id, args, kwargs
        )

    async def launch_fire_and_forget_async_job(
        self, qualified_name: str, module_id: str, args=None, kwargs=None
    ):
        return await self._launch_job(
            LaunchJobMode.ASYNC_FIRE_AND_FORGET, qualified_name, module_id, args, kwargs
        )

    def _build_request(
        self, mode: LaunchJobMode, qualified_name: str, module_id: str, args, kwargs
    ):
        encoded_args = b""
        if args or kwargs:
            encoded_args = codec.encode_args(
                args if args else None,
                kwargs if kwargs else None,
            ).encode("utf-8")

        job = remote_pb2.RemoteJob(
            container_image=symbols.find_image_for_module(module_id),
            qualified_symbol=qualified_name,
            namespace=namespace.get(),
            encoded_args=encoded_args,
            module_id=module_id,
        )
        request: Union[remote_pb2.LaunchJobRequest, remote_pb2.LaunchBlockingJobRequest]

        if mode.is_fire_and_forget():
            request = remote_pb2.LaunchJobRequest(job=job)
        elif mode.is_wait_for_response():
            request = remote_pb2.LaunchBlockingJobRequest(job=job)
        else:
            raise Exception(f"unsupported mode: {mode}")

        if self.dry_run:
            self.dry_run_last_request = request

        return request

    def _launch_job(
        self, mode: LaunchJobMode, qualified_name: str, module_id: str, args, kwargs
    ):
        request = self._build_request(mode, qualified_name, module_id, args, kwargs)

        # TODO(Landau): This is gross, remove dry_run.
        if self.dry_run:
            if mode.is_async():

                async def empty_awaitable():
                    pass

                return empty_awaitable()
            return

        # If dialing is slow, consider dialing earlier.
        if not self.is_dialed:
            self.dial()

        if mode == LaunchJobMode.FIRE_AND_FORGET:
            return self.stub.LaunchJob(request)
        elif mode == LaunchJobMode.BLOCKING:
            return self.stub.LaunchBlockingJob(request)

        # Async flavors return a coroutine.
        elif mode == LaunchJobMode.ASYNC:
            return self.async_stub.LaunchBlockingJob(request)
        elif mode == LaunchJobMode.ASYNC_FIRE_AND_FORGET:
            return self.async_stub.LaunchJob(request)
        else:
            raise Exception(f"unsupported mode: {mode}")

    def _transform_response_exception(self, response):
        if not isinstance(response, remote_pb2.LaunchBlockingJobResponse):
            # can happen for dry-run
            # TODO this is a code-smell. We should instead properly mock
            # the response from the grpc call.
            return

        if (
            response.result.error is None
            or response.result.error.code == remote_pb2.ErrorCode.UNKNOWN
        ):
            return response.result.value
        else:
            if response.result.error.code == remote_pb2.ErrorCode.APPLICATION:
                raise ApplicationException(response.result.error.message)
            else:
                raise RuntimeException(response.result.error.message)



from typing import Callable

from jetpack import _job
from jetpack._job import client as _client
from jetpack._job.job import Job as _Job
from jetpack._remote.codec import FuncArgs


class JobDecorator:
    def __call__(self, fn: Callable) -> Callable:
        return _Job(fn)


job = JobDecorator()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jetpack/proto/runtime/v1alpha1/remote.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='jetpack/proto/runtime/v1alpha1/remote.proto',
  package='remote',
  syntax='proto3',
  serialized_options=b'\n\ncom.remoteB\013RemoteProtoP\001Z2go.jetpack.io/proto/jetpack/proto/runtime/v1alpha1\242\002\003RXX\252\002\006Remote\312\002\006Remote\342\002\022Remote\\GPBMetadata\352\002\006Remote',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n+jetpack/proto/runtime/v1alpha1/remote.proto\x12\x06remote\"[\n\x11RemoteCallRequest\x12)\n\x10qualified_symbol\x18\x01 \x01(\tR\x0fqualifiedSymbol\x12\x1b\n\tjson_args\x18\x02 \x01(\tR\x08jsonArgs\"7\n\x12RemoteCallResponse\x12!\n\x0cjson_results\x18\x01 \x01(\tR\x0bjsonResults\"7\n\x10LaunchJobRequest\x12#\n\x03job\x18\x01 \x01(\x0b\x32\x11.remote.RemoteJobR\x03job\"\x13\n\x11LaunchJobResponse\"?\n\x18LaunchBlockingJobRequest\x12#\n\x03job\x18\x01 \x01(\x0b\x32\x11.remote.RemoteJobR\x03job\"C\n\x19LaunchBlockingJobResponse\x12&\n\x06result\x18\x01 \x01(\x0b\x32\x0e.remote.ResultR\x06result\"\xbd\x01\n\tRemoteJob\x12\'\n\x0f\x63ontainer_image\x18\x01 \x01(\tR\x0e\x63ontainerImage\x12)\n\x10qualified_symbol\x18\x02 \x01(\tR\x0fqualifiedSymbol\x12\x1c\n\tnamespace\x18\x03 \x01(\tR\tnamespace\x12!\n\x0c\x65ncoded_args\x18\x04 \x01(\x0cR\x0b\x65ncodedArgs\x12\x1b\n\tmodule_id\x18\x05 \x01(\tR\x08moduleId\"R\n\x06Result\x12#\n\x05value\x18\x01 \x01(\x0b\x32\r.remote.ValueR\x05value\x12#\n\x05\x65rror\x18\x02 \x01(\x0b\x32\r.remote.ErrorR\x05\x65rror\"m\n\x05\x45rror\x12%\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x11.remote.ErrorCodeR\x04\x63ode\x12\x18\n\x07message\x18\x02 \x01(\tR\x07message\x12#\n\rencoded_error\x18\x03 \x01(\x0cR\x0c\x65ncodedError\",\n\x05Value\x12#\n\rencoded_value\x18\x01 \x01(\x0cR\x0c\x65ncodedValue*7\n\tErrorCode\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07RUNTIME\x10\x64\x12\x10\n\x0b\x41PPLICATION\x10\xc8\x01\x32\xf7\x01\n\x0eRemoteExecutor\x12\x45\n\nRemoteCall\x12\x19.remote.RemoteCallRequest\x1a\x1a.remote.RemoteCallResponse\"\x00\x12\x42\n\tLaunchJob\x12\x18.remote.LaunchJobRequest\x1a\x19.remote.LaunchJobResponse\"\x00\x12Z\n\x11LaunchBlockingJob\x12 .remote.LaunchBlockingJobRequest\x1a!.remote.LaunchBlockingJobResponse\"\x00\x42\x85\x01\n\ncom.remoteB\x0bRemoteProtoP\x01Z2go.jetpack.io/proto/jetpack/proto/runtime/v1alpha1\xa2\x02\x03RXX\xaa\x02\x06Remote\xca\x02\x06Remote\xe2\x02\x12Remote\\GPBMetadata\xea\x02\x06Remoteb\x06proto3'
)

_ERRORCODE = _descriptor.EnumDescriptor(
  name='ErrorCode',
  full_name='remote.ErrorCode',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RUNTIME', index=1, number=100,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPLICATION', index=2, number=200,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=850,
  serialized_end=905,
)
_sym_db.RegisterEnumDescriptor(_ERRORCODE)

ErrorCode = enum_type_wrapper.EnumTypeWrapper(_ERRORCODE)
UNKNOWN = 0
RUNTIME = 100
APPLICATION = 200



_REMOTECALLREQUEST = _descriptor.Descriptor(
  name='RemoteCallRequest',
  full_name='remote.RemoteCallRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='qualified_symbol', full_name='remote.RemoteCallRequest.qualified_symbol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='qualifiedSymbol', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='json_args', full_name='remote.RemoteCallRequest.json_args', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='jsonArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=55,
  serialized_end=146,
)


_REMOTECALLRESPONSE = _descriptor.Descriptor(
  name='RemoteCallResponse',
  full_name='remote.RemoteCallResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='json_results', full_name='remote.RemoteCallResponse.json_results', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='jsonResults', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=203,
)


_LAUNCHJOBREQUEST = _descriptor.Descriptor(
  name='LaunchJobRequest',
  full_name='remote.LaunchJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='job', full_name='remote.LaunchJobRequest.job', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='job', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=205,
  serialized_end=260,
)


_LAUNCHJOBRESPONSE = _descriptor.Descriptor(
  name='LaunchJobResponse',
  full_name='remote.LaunchJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=281,
)


_LAUNCHBLOCKINGJOBREQUEST = _descriptor.Descriptor(
  name='LaunchBlockingJobRequest',
  full_name='remote.LaunchBlockingJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='job', full_name='remote.LaunchBlockingJobRequest.job', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='job', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=283,
  serialized_end=346,
)


_LAUNCHBLOCKINGJOBRESPONSE = _descriptor.Descriptor(
  name='LaunchBlockingJobResponse',
  full_name='remote.LaunchBlockingJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='remote.LaunchBlockingJobResponse.result', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='result', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=348,
  serialized_end=415,
)


_REMOTEJOB = _descriptor.Descriptor(
  name='RemoteJob',
  full_name='remote.RemoteJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='container_image', full_name='remote.RemoteJob.container_image', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='containerImage', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qualified_symbol', full_name='remote.RemoteJob.qualified_symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='qualifiedSymbol', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='remote.RemoteJob.namespace', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='namespace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encoded_args', full_name='remote.RemoteJob.encoded_args', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='encodedArgs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module_id', full_name='remote.RemoteJob.module_id', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='moduleId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=418,
  serialized_end=607,
)


_RESULT = _descriptor.Descriptor(
  name='Result',
  full_name='remote.Result',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='remote.Result.value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='remote.Result.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='error', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=609,
  serialized_end=691,
)


_ERROR = _descriptor.Descriptor(
  name='Error',
  full_name='remote.Error',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='remote.Error.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='code', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='remote.Error.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='message', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='encoded_error', full_name='remote.Error.encoded_error', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='encodedError', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=693,
  serialized_end=802,
)


_VALUE = _descriptor.Descriptor(
  name='Value',
  full_name='remote.Value',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='encoded_value', full_name='remote.Value.encoded_value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='encodedValue', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=804,
  serialized_end=848,
)

_LAUNCHJOBREQUEST.fields_by_name['job'].message_type = _REMOTEJOB
_LAUNCHBLOCKINGJOBREQUEST.fields_by_name['job'].message_type = _REMOTEJOB
_LAUNCHBLOCKINGJOBRESPONSE.fields_by_name['result'].message_type = _RESULT
_RESULT.fields_by_name['value'].message_type = _VALUE
_RESULT.fields_by_name['error'].message_type = _ERROR
_ERROR.fields_by_name['code'].enum_type = _ERRORCODE
DESCRIPTOR.message_types_by_name['RemoteCallRequest'] = _REMOTECALLREQUEST
DESCRIPTOR.message_types_by_name['RemoteCallResponse'] = _REMOTECALLRESPONSE
DESCRIPTOR.message_types_by_name['LaunchJobRequest'] = _LAUNCHJOBREQUEST
DESCRIPTOR.message_types_by_name['LaunchJobResponse'] = _LAUNCHJOBRESPONSE
DESCRIPTOR.message_types_by_name['LaunchBlockingJobRequest'] = _LAUNCHBLOCKINGJOBREQUEST
DESCRIPTOR.message_types_by_name['LaunchBlockingJobResponse'] = _LAUNCHBLOCKINGJOBRESPONSE
DESCRIPTOR.message_types_by_name['RemoteJob'] = _REMOTEJOB
DESCRIPTOR.message_types_by_name['Result'] = _RESULT
DESCRIPTOR.message_types_by_name['Error'] = _ERROR
DESCRIPTOR.message_types_by_name['Value'] = _VALUE
DESCRIPTOR.enum_types_by_name['ErrorCode'] = _ERRORCODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RemoteCallRequest = _reflection.GeneratedProtocolMessageType('RemoteCallRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOTECALLREQUEST,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteCallRequest)
  })
_sym_db.RegisterMessage(RemoteCallRequest)

RemoteCallResponse = _reflection.GeneratedProtocolMessageType('RemoteCallResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOTECALLRESPONSE,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteCallResponse)
  })
_sym_db.RegisterMessage(RemoteCallResponse)

LaunchJobRequest = _reflection.GeneratedProtocolMessageType('LaunchJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHJOBREQUEST,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchJobRequest)
  })
_sym_db.RegisterMessage(LaunchJobRequest)

LaunchJobResponse = _reflection.GeneratedProtocolMessageType('LaunchJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHJOBRESPONSE,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchJobResponse)
  })
_sym_db.RegisterMessage(LaunchJobResponse)

LaunchBlockingJobRequest = _reflection.GeneratedProtocolMessageType('LaunchBlockingJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHBLOCKINGJOBREQUEST,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchBlockingJobRequest)
  })
_sym_db.RegisterMessage(LaunchBlockingJobRequest)

LaunchBlockingJobResponse = _reflection.GeneratedProtocolMessageType('LaunchBlockingJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHBLOCKINGJOBRESPONSE,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchBlockingJobResponse)
  })
_sym_db.RegisterMessage(LaunchBlockingJobResponse)

RemoteJob = _reflection.GeneratedProtocolMessageType('RemoteJob', (_message.Message,), {
  'DESCRIPTOR' : _REMOTEJOB,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteJob)
  })
_sym_db.RegisterMessage(RemoteJob)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Result)
  })
_sym_db.RegisterMessage(Result)

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
  'DESCRIPTOR' : _ERROR,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Error)
  })
_sym_db.RegisterMessage(Error)

Value = _reflection.GeneratedProtocolMessageType('Value', (_message.Message,), {
  'DESCRIPTOR' : _VALUE,
  '__module__' : 'jetpack.proto.runtime.v1alpha1.remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Value)
  })
_sym_db.RegisterMessage(Value)


DESCRIPTOR._options = None

_REMOTEEXECUTOR = _descriptor.ServiceDescriptor(
  name='RemoteExecutor',
  full_name='remote.RemoteExecutor',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=908,
  serialized_end=1155,
  methods=[
  _descriptor.MethodDescriptor(
    name='RemoteCall',
    full_name='remote.RemoteExecutor.RemoteCall',
    index=0,
    containing_service=None,
    input_type=_REMOTECALLREQUEST,
    output_type=_REMOTECALLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LaunchJob',
    full_name='remote.RemoteExecutor.LaunchJob',
    index=1,
    containing_service=None,
    input_type=_LAUNCHJOBREQUEST,
    output_type=_LAUNCHJOBRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LaunchBlockingJob',
    full_name='remote.RemoteExecutor.LaunchBlockingJob',
    index=2,
    containing_service=None,
    input_type=_LAUNCHBLOCKINGJOBREQUEST,
    output_type=_LAUNCHBLOCKINGJOBRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_REMOTEEXECUTOR)

DESCRIPTOR.services_by_name['RemoteExecutor'] = _REMOTEEXECUTOR

# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import base64
import inspect
import os
from typing import Callable, Dict, Optional, ValuesView

from jetpack import utils
from jetpack._job import client as _client
from jetpack._job import interface
from jetpack._remote import codec
from jetpack.proto.runtime.v1alpha1 import remote_pb2


class AlreadyExistsError(Exception):
    pass


class Job:
    # Question: do we need a separate Dict for blocking jobs?
    jobs: Dict[str, Job] = {}

    def __init__(self, func: Callable):
        self.func = func
        Job.add_job(self)

    def __call__(self, *args, **kwargs):
        """
        No return value for now, but eventually this will return a handle
        or an awaitable

        For now only self module is allowed. JETPACK_MODULE_ID must be set.
        """
        module_id = os.environ["JETPACK_MODULE_ID"]
        if inspect.iscoroutinefunction(self.func):
            return _client.launch_async_job(self.name(), module_id, args, kwargs)
        else:
            return _client.launch_job(self.name(), module_id, args, kwargs)

    def fire_and_forget(self, *args, **kwargs):
        module_id = os.environ["JETPACK_MODULE_ID"]
        if inspect.iscoroutinefunction(self.func):
            return _client.launch_fire_and_forget_async_job(
                self.name(), module_id, args, kwargs
            )
        else:
            return _client.launch_fire_and_forget_job(
                self.name(), module_id, args, kwargs
            )

    def name(self) -> str:
        return utils.qualified_func_name(self.func)

    def exec(self, base64_encoded_args: str = "") -> remote_pb2.RemoteCallResponse:

        args, kwargs = None, None
        if base64_encoded_args:
            encoded_args = base64.b64decode(base64_encoded_args).decode("utf-8")
            args, kwargs = codec.decode_args(encoded_args)

        # ensure we default since encoded args might have None in them
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        if inspect.iscoroutinefunction(self.func):
            result = asyncio.run(self.func(*args, **kwargs))
        else:
            result = self.func(*args, **kwargs)
        # for now just return encoded result, but we could send it to a redis channel
        # that the caller might be listening to.
        return codec.encode_result(result)

    @classmethod
    def add_job(cls, job: Job):
        if job.name() in cls.jobs:
            raise AlreadyExistsError(f"Job with name {job.name()} already exists")
        cls.jobs[job.name()] = job

    @classmethod
    def find_job(cls, name: str) -> Optional[Job]:
        return cls.jobs.get(name)

    @classmethod
    def defined_jobs(cls) -> ValuesView[Job]:
        return cls.jobs.values()
